/*
* This file is part of Firestorm NIDS
* Copyright (c) 2001 Gianni Tedesco
*
*    Firestorm is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    Firestorm is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with Firestorm; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <matcher.h>
#include <s_util.h>
#include <cleanup.h>
#include <targets.h>

/* List of registered matchers   */
p_matcher	matches=NULL;
mlock_t		matches_lock;

/* List of all added rules */
p_rule		rules=NULL;
mlock_t		rules_lock;

p_vertex	rule_tree=NULL;
mlock_t		rtree_lock;
uint		k_value=0;

uint		num_pkt=0;
uint		num_match=0;

int matcher_add_rule(p_rule r)
{
	if ( r == NULL ) return 0;

	/* FIXME: validate rule better */
	if ( r->criteria == NULL ) return 0;

	CRITICAL_START(matches_lock);
	rules=list_append(r, rules);
	globals.num_rules++;
	CRITICAL_STOP(matches_lock);

	return 1;
}

int matcher_set_rule_target(p_rule r, char *target, char *val)
{
	p_target t;
	
	if ( !r || !target ) return 0;

	if ( !(t=target_get(target)) ) return 0;

	if ( !(r->t_priv=t->validate(val)) ) return 0;

	r->target=t;
	
	return 1;
}

/* Add a criteria to a rule */
int matcher_add_criteria_to_rule(
	p_rule r, 
	char *crit,
	char *val,
	p_kv opts,
	int negate)
{
	p_matcher matcher=NULL;
	p_relement rel=NULL;
	
	if ( !r || !crit || !val ) return 0;

	if ( (matcher=matcher_get(crit)) == NULL ) {
		print_out(PRINT_ERROR "Unable to add rule, no such "
			"matcher (%s)" CRLF, crit);
		return 0;
	}

	if ( (rel=globals.malloc(sizeof(relement_t))) == NULL  ) {
		return 0;
	}

	/* Validate this rule criteria */
	rel->match = matcher;
	if ( (rel->priv=rel->match->validate(val, opts)) == NULL ) {
		globals.free(rel);
		return 0;
	}

	rel->negate = (negate) ? 1 : 0;
	
	/* Now add the criteria to the rule */
	r->criteria = list_append(rel, r->criteria);

	globals.num_ops++;
	return 1;
}

/* 
* Cleans up a rule object (MUST have been
* safely initialised with SAFE_INIT_RULE_PTR() 
* before calling this
*/
void matcher_rule_cleanup(p_rule r)
{
	p_relement r_foo, r_bar;

	if ( r->target && r->t_priv )
		r->target->cleanup(r->t_priv);
	
	for (r_foo=r->criteria; r_foo;) {
		r_bar=r_foo;
		r_foo=(p_relement)r_foo->l.next;
		globals.free(r_bar);
	}

	globals.free(r);
}

void matcher_cleanup(int code, void *priv)
{
	p_rule foo,bar;
	p_matcher m_foo, m_bar;
	
	print_raw("Removing all rules" CRLF);
	CRITICAL_START(rules_lock);
	for (foo=rules; foo;) {
		bar=foo;
		foo=(p_rule)foo->l.next;
		matcher_rule_cleanup(bar);
	}
	CRITICAL_STOP(rules_lock);
	
	print_raw("Removing all matchers" CRLF);
	CRITICAL_START(matches_lock);
	for (m_foo=matches; m_foo;) {
		m_bar=m_foo;
		m_foo=(p_matcher)m_foo->l.next;
		globals.free(m_bar);
	}
	CRITICAL_STOP(matches_lock);
}

int matcher_init(void)
{
	MUTEX_CREATE(matches_lock, "Matches");
	cleanup_add_handler(cleanup_lock, (void *)&matches_lock);

	MUTEX_CREATE(rules_lock, "Rules");
	cleanup_add_handler(cleanup_lock, (void *)&rules_lock);

	MUTEX_CREATE(rtree_lock, "Rule tree");
	cleanup_add_handler(cleanup_lock, (void *)&rtree_lock);
	
	cleanup_add_handler(matcher_cleanup, NULL);
	
	return 1;
}

/* Get a matcher by its name  */
p_matcher matcher_get(char *n)
{
	p_matcher tmp;

	for ( tmp=matches; tmp; tmp=(p_matcher)tmp->l.next ) {
		if ( strcasecmp(tmp->name, n) == 0 ) {
			return tmp;
		}
	}

	return NULL;
}

/* Plugins call this to add a new matcher */
int matcher_add ( char *n, proc_match_validate v, 
		proc_match_cleanup cl,
		proc_match_match match,
		proc_match_compare compare)
{
	p_matcher foo;

	if ( n == NULL ) return 0;
	if ( v == NULL ) return 0;
	if ( cl == NULL ) return 0;
	if ( compare == NULL ) return 0;
	
	if ( matcher_get(n) ) {
		print_out(PRINT_WARNING 
			"Tried to load %s matcher twice" CRLF, n);
		return 0;
	}
	
	if ( !(foo=globals.malloc(sizeof(matcher_t))) ) {
		return 0;
	}

	foo->name=n;
	foo->validate=v;
	foo->cleanup=cl;
	foo->match=match;
	foo->compare=compare;
	
	CRITICAL_START(matches_lock);
	matches=list_append(foo, matches);
	globals.num_matchers++;
	CRITICAL_STOP(matches_lock);

	return 1;
}

/* Make a unique list of vertexes from a list of criteria */
p_vertex matcher_unique(p_vertex u, p_relement c)
{
	p_vertex	tmp,n;
	p_vertex	ret=u;
	int		is_unique;
	p_relement	t;

	for (t=c; t; t=(p_relement)t->l.next) {
		is_unique=1;

		for(tmp=ret; tmp; tmp=(p_vertex)tmp->l.next) {
			if ( (t->match==tmp->match) &&
				(t->match->compare(t->priv, tmp->priv))) {
				is_unique=0;
				tmp->usage++;

				/* Replace duplicate with pointer to original */
				t->match->cleanup(t->priv);
				t->priv=tmp->priv;
				break;
			}
		}

		if ( is_unique ) {
			if ( (n=globals.malloc(sizeof(vertex_t))) ) {
				memset(n,0,sizeof(vertex_t));
				n->priv=t->priv;
				n->match=t->match;
				n->usage=1;
				ret=list_append(n,ret);
			}else{
				print_out(PRINT_ERROR "Not enough mem for tree vertex");
			}
		}
	}
	
	return ret;
}

/* Adds a child on to a tree vertex */
void matcher_add_child(p_vertex root, p_vertex v, int child, p_relement rule)
{
	p_vertex n;
	
	/* Check it doesnt already have a child */
	if ( v->child[child] ) return;

	for(n=root; n; n=(p_vertex)n->l.next) {
		if ( (n->match==rule->match) && (n->priv==rule->priv) ) {
			v->child[child]=n;
			print_out("%s.child[%i] = 0x%x" CRLF,
				v->match->name, child, n);
			break;
		}
	}
}

uint matcher_rule_to_tree(p_vertex root, p_vertex t, p_relement c, p_rule a)
{
	p_relement tmp;
	p_relement us=NULL,nxt=NULL;
	int cur;
	int ret=1, ret1, ret2;
	p_vertex n;

	if ( !root ) {print_out("root cant be NULL" CRLF); return 0;}
	if ( !t ) {print_out("t cant be NULL" CRLF); return 0;}
	if ( !c ) {print_out("c cant be NULL" CRLF); return 0;}
	if ( !a ) {print_out("a cant be NULL" CRLF); return 0;}

	/* Check for metadata criteria */
	if ( !c->match->match ) {
		if ( c->l.next ) 
			ret+=matcher_rule_to_tree(root, t, (p_relement)c->l.next, a);
		return ret;
	}
	
	/* Try find the right criteria for this vertex */
	if ( (c->match==t->match) && (c->priv==t->priv) ) {
		us=c;
		nxt=(p_relement)c->l.next;
	}else{
		for(tmp=a->criteria; tmp; tmp=(p_relement)tmp->l.next) {
			if ( (tmp->match==t->match)&&(tmp->priv==t->priv)) {
				us=tmp;
				break;
			}
		}
		nxt=c;
	}

	if ( us ) { /* Descend down only one side of this vertex */
		cur=(us->negate) ? 0 : 1;
		if ( nxt ) {
			matcher_add_child(root, t, cur, nxt);
			ret+=matcher_rule_to_tree(root, t->child[cur], nxt, a);
		}else{
			/* End of the line */
			print_out("%s.action[%i] = %s" CRLF,
				t->match->name, cur,
				(a->t_priv==0xdeadbeef) ? "<DROP>":a->t_priv);
			t->action[cur]=a;
		}
	}else{ /* Descend both sides */
		if ( !nxt->l.next ) {
			print_out("**FIXME: Loop detected, no workaround!" CRLF);
			return ret;
		}
			
		if ( !t->child[0] ) matcher_add_child(root, t, 0, nxt);
		ret1=matcher_rule_to_tree(root, t->child[0], nxt, a);
		
		if ( !t->child[1] ) matcher_add_child(root, t, 1, nxt);
		ret2=matcher_rule_to_tree(root, t->child[1], nxt, a);
		
		ret+=(ret1>ret2) ? ret1 : ret2;
	}
	return ret;
}

int matcher_compile(void)
{
	p_rule r;
	p_vertex v;
	int k;
	
	print_out("Compiling rules..." CRLF);

	/* Make a list of unique tree vertexes */
	for (r=rules; r; r=(p_rule)r->l.next) {
		rule_tree=matcher_unique(rule_tree, r->criteria);
	}

	/* TODO: Sort list */
	
	/* print the list of vertexes */
#if 1
	for ( v=rule_tree; v; v=(p_vertex)v->l.next ) {
		print_out("0x%x: usage=%u priv=0x%x\tmatch=%s" CRLF,
			v, v->usage, v->priv, v->match->name);
	}
#endif

	/* Generate the tree */
	k_value=0;
	for(r=rules; r; r=(p_rule)r->l.next) {
		print_out(CRLF "<-- Adding rule" CRLF);
		k=matcher_rule_to_tree(rule_tree, rule_tree, r->criteria, r);
		k_value=(k>k_value) ? k : k_value;
	}
	
	print_out("Compilation complete (k=%u)." CRLF, k_value);
	return 1;
}

/* Actually match a packet agains all our rules */
void matcher_begin_iterative( p_packet p )
{	
	int m=1;
	p_rule r=NULL;
	p_relement c=NULL;
	char *def_target;
	p_target tgt;
	
	if ( rules == NULL ) {
		if ( !(def_target=plugin_get_var("DEFAULT_TARGET")) ) {
			def_target="ALERT";
		}
		
		if ( (tgt=target_get(def_target)) )
			tgt->target(p, "PACKET RECIEVED", rules);
		else
			print_out("$DEFAULT_TARGET(%s) not loaded, and no rules" CRLF, def_target);
		return;
	}

	for (r=rules; r; r=(p_rule)r->l.next) {
		m=1;
		
		for (c=r->criteria; c; c=(p_relement)c->l.next) {
			/* Skip meta-data nodes */
			if ( c->match->match == NULL ) continue;
	
			num_match++;		
			if ( !(c->match->match(c->priv, p, c->negate)) ){
				m=0;
				break;
			}
		}
		
		if (m) {
			/* Send the packet to the rule target */
			switch ( r->target->target(p, r->t_priv, r) ) {
				case TARGET_DONE:
					return;
				case TARGET_CONTINUE:
				default:
					break;	/* continue */
			}
		}
	}
}

int matcher_begin_tree(p_vertex root, p_packet p)
{
	uint x;

	if ( !root ) return 0;
	if ( !root->match ) return 0;
	if ( !root->match->match ) return 0;

	x=root->match->match(root->priv, p, 0);
	if ( x > 1 ) x==1;
	
	num_match++;
	if ( root->action[x] ) {
		root->action[x]->target->target(p,
			root->action[x]->t_priv,
			root->action[x]);
	}

	if ( root->child[x] ) {
		return matcher_begin_tree(root->child[x], p);
	}else{
		return 0;
	}
}

void matcher_stats(void)
{
	print_out("MATCHER STATS: %u packets" CRLF, num_pkt);
	print_out("MATCHER STATS: %u operations" CRLF, num_match);
	print_out("MATCHER_STATS: %u ops/pkt" CRLF, num_match/num_pkt);
}

void matcher_begin( p_packet p )
{
	num_pkt++;

#ifdef _COMPILE_RULES
	matcher_begin_tree(rule_tree, p);
#else
	matcher_begin_iterative(p);
#endif
}
