/*
* This file is part of Firestorm NIDS
* Copyright (c) 2001 Gianni Tedesco
*
*    Firestorm is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    Firestorm is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with Firestorm; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <report.h>
#include <s_util.h>
#include <zlib.h>

PLUGIN_DECLARE_STD_IMPORTS;
proc_target_add		add_target;

int udp_fd=-1;	/* All share one socket to be nice */
mlock_t fdl;	/* Lock for the socket */
struct sockaddr_in svr;

u_int32_t compress_zpacket(p_fstp_pkt p, int level)
{
	char buf[FSTP_MAX_PAYLOAD];
	int dlen=FSTP_MAX_PAYLOAD;

	/* Copy data in to the temp buffer */
	memcpy((char *)&buf, (char *)&p->data, p->hdr.len-FSTP_HDR_SIZE);

	/* Compress it back in to the original packet */
	compress2((char *)&p->data, &dlen, (char *)&buf, p->hdr.len-FSTP_HDR_SIZE, level);

	return FSTP_HDR_SIZE+dlen;
}

void add_padding(p_fstp_pkt pkt, u_int32_t len)
{
	void *ptr=pkt;
	ptr += pkt->hdr.len;

	/* Add the header */
	((p_segment_hdr)ptr)->type=FSTP_SEG_PADDING;
	((p_segment_hdr)ptr)->caplen=len;
	
	/* Copy the data */
	ptr+=FSTP_SEG_SIZE;
	memset(ptr, 0x00, len);
	
	/* Update the length */
	pkt->hdr.len += len + FSTP_SEG_SIZE;
}

void send_packet(p_fstp_pkt pkt)
{
	if ( !pkt->hdr.len ) return;

#if 0
	if ( pkt->hdr.len + FSTP_SEG_SIZE < FSTP_MAX_PAYLOAD ) {
		add_padding(FSTP_MAX_PAYLOAD-FSTP_SEG_SIZE-pkt->hdr.len);
	}
#endif

	/* Check the canary word (catch internal errors) */
	if ( pkt->canary != FSTP_CANARY ) {
		print_out("Warning canary == 0x%x" CRLF, pkt->canary);
	}

	/* Add the EOT segment */
	((p_segment_hdr)(((void*)pkt)+pkt->hdr.len))->type=FSTP_SEG_EOT;
	((p_segment_hdr)(((void*)pkt)+pkt->hdr.len))->caplen=0;
	pkt->hdr.len+=FSTP_EOT_SIZE;
	
	/* Try some compression */
	if ( pkt->hdr.len > 512 ) {
		pkt->hdr.compression = FSTP_COMPRESS_ZLIB;
		pkt->hdr.level = Z_BEST_COMPRESSION;
		pkt->hdr.len=compress_zpacket(pkt, Z_BEST_COMPRESSION);
	}else{
		pkt->hdr.compression = FSTP_COMPRESS_NONE;
		pkt->hdr.level = 0;
	}

	pkt->hdr.hdr_len = FSTP_HDR_SIZE;
	pkt->hdr.csum = 0; /* Need a nice 32bit crypto hash */

	/* Actually send the packet */
	CRITICAL_START(fdl);
	sendto(udp_fd, (char *)pkt, pkt->hdr.len,
		0, (struct sockaddr *)&svr, sizeof(svr));
	CRITICAL_STOP(fdl);

	pkt->hdr.len=FSTP_HDR_SIZE;
	pkt->canary=FSTP_CANARY;
}

void add_segment(p_fstp_pkt pkt, p_packet p, u_int32_t len)
{
	void *ptr=pkt;
	ptr += pkt->hdr.len;

	/* Add the header */
	((p_segment_hdr)ptr)->type=FSTP_SEG_PACKET;
	((p_segment_hdr)ptr)->caplen=len;
	
	/* Copy the data */
	ptr+=FSTP_SEG_SIZE;
	memcpy(ptr, (char *)&p->data, len);
	
	/* Update the length */
	pkt->hdr.len += len + FSTP_SEG_SIZE;
}

target_action_e report_target(p_packet p, void *priv, p_rule r)
{
	p_report_priv ps=priv;

	CRITICAL_START(ps->plock);
	
	/* Check we can fit in to this packet */
	if (p->caplen+ps->pkt.hdr.len+FSTP_SEG_SIZE+FSTP_EOT_SIZE >= FSTP_MAX_PAYLOAD) {
		send_packet(&ps->pkt);
	}

	if ( p->caplen + FSTP_SEG_SIZE + FSTP_EOT_SIZE > FSTP_MAX_PAYLOAD ) {
		send_packet(&ps->pkt);
		add_segment(&ps->pkt, p, FSTP_MAX_PAYLOAD-FSTP_SEG_SIZE);
		print_out(PRINT_WARNING "Truncated %i byte packet to %i bytes!" CRLF,
			p->caplen, FSTP_MAX_PAYLOAD-FSTP_SEG_SIZE);
	}else{
		add_segment(&ps->pkt, p, p->caplen);
	}
	
	CRITICAL_STOP(ps->plock);
	return TARGET_DONE;
}

void *report_validate(char *token)
{
	p_report_priv x;

	if ( !(x=my_malloc(sizeof(report_priv_t))) ) {
		return NULL;
	}

	x->pkt.canary=FSTP_CANARY;
	x->pkt.hdr.len=sizeof(x->pkt.hdr);
	MUTEX_CREATE(x->plock, "Individual report packet header");
	return x;
}

void report_cleanup(void *priv)
{
	p_report_priv x;

	if ( !(x=(p_report_priv)priv) ) return;

	if ( x->pkt.hdr.len > sizeof(x->pkt.hdr)) {
		CRITICAL_START(x->plock);
		send_packet(&x->pkt);
		CRITICAL_STOP(x->plock);
	}

	MUTEX_DESTROY(x->plock);
}

uint _PLUGIN_INIT (p_plugin_in in, p_plugin_out out)
{
	char *prt, *ip;
	struct hostent *h;
	struct in_addr a;
	PLUGIN_CHECK_PARAMS(in);
	PLUGIN_IMPORT_STD;
	
	add_target=in->add_target;
	
	PLUGIN_NAME("target/report");
	PLUGIN_DESC("Reports to server");
	PLUGIN_VERS(1,0);
	
	MUTEX_CREATE(fdl, "Report socket lock");
	
	/* For now the server is hardcoded 127.0.0.1:2233 */
	udp_fd=socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);

	if ( (prt=get_var("FSTP_PORT")) ) {
		svr.sin_port=htons(atoi(prt));
		if ( svr.sin_port == 0 ) {
			print_out(PRINT_ERROR "What kind of port is $FSTP_PORT(%s)?" CRLF, prt);
			return PLUGIN_ERR_FAIL;
		}
	}else{
		svr.sin_port=htons(FSTP_DEFAULT_PORT);
	}

	if ( (ip=get_var("FSTP_SERVER")) ) {
		if ( inet_aton(ip, &a) ) {
			svr.sin_addr.s_addr=a.s_addr;
		}else if ( (h=gethostbyname(ip)) ) {
			memcpy(&a, h->h_addr, h->h_length);
			svr.sin_addr.s_addr=a.s_addr;
		}else{
			print_out(PRINT_ERROR "Couldn't find address $FSTP_SERVER(%s)" CRLF, ip);
			return PLUGIN_ERR_FAIL;
		}
	}else{
		svr.sin_addr.s_addr=htonl(INADDR_LOOPBACK);
	}
	
	/*
	* TODO: Startup a maintenance thread which keeps a linked
	* list of all the private structures and periodically
	* flushes them. 
	* */
	
	add_target("REPORT",
		report_target,
		report_validate,
		report_cleanup);
	
	return PLUGIN_ERR_OK;
}

/* This is where we would cleanup our shit */
uint _PLUGIN_UNLOAD (void)
{
	if ( udp_fd >= 0 ) close(udp_fd);
	MUTEX_DESTROY(fdl);
	return PLUGIN_ERR_OK;
}
